import openpaperwork_core

from .... import _
from . import BaseDocViewController


class TitleController(BaseDocViewController):
    def enter(self):
        folded = self.plugin.core.call_success("mainwindow_get_folded")
        if folded:
            self.plugin.widget_tree.get_object(
                "docview_header"
            ).set_title("")
            return

        (doc_id, doc_url) = self.plugin.active_doc
        if self.plugin.core.call_success("is_doc", doc_url) is not None:
            doc_date = self.plugin.core.call_success(
                "doc_get_date_by_id", doc_id
            )
            if doc_date is not None:
                doc_date = self.plugin.core.call_success(
                    "i18n_date_short", doc_date
                )
            else:
                doc_date = doc_id
            self.plugin.widget_tree.get_object(
                "docview_header"
            ).set_title(doc_date)
        else:
            self.plugin.widget_tree.get_object(
                "docview_header"
            ).set_title(_("New document"))

    def on_close(self):
        self.plugin.widget_tree.get_object("docview_header").set_title("")


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.controller = None

    def get_interfaces(self):
        return ['gtk_docview_controller']

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def gtk_docview_get_controllers(self, out: dict, docview):
        self.controller = TitleController(docview)
        out['title'] = self.controller

    def on_mainwindow_fold_change(self):
        if self.controller is None:
            return
        # update the title bar
        self.controller.enter()



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.windows = []

    def get_interfaces(self):
        return [
            'gtk_about',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'authors',
                'defaults': ['paperwork_backend.authors'],
            },
            {
                'interface': 'app',
                'defaults': ['paperwork_backend.app'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'icon',
                'defaults': ['paperwork_gtk.icon'],
            },
        ]

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def _set_authors_str(self, method, authors):
        out = ""
        for (email, name, line_count) in authors:
            if line_count > 0:
                out += "{} ({})\n".format(name, line_count)
            else:
                out += "{}\n".format(name)
        method(out)

    def _set_authors_list(self, method, authors):
        out = []
        for (email, name, line_count) in authors:
            txt = name
            if line_count > 0:
                txt += " ({})".format(line_count)
            out.append(txt)
        method(out)

    def _on_close(self, dialog, *args, **kwargs):
        LOGGER.info("Closing dialog 'about'")
        dialog.destroy()
        self.core.call_all("on_gtk_window_closed", dialog)

    def gtk_open_about(self):
        LOGGER.info("Opening dialog 'about'")
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.about", "about.glade"
        )

        authors = {}
        self.core.call_all("authors_get", authors)

        documentation = authors.pop("Documentation", [])
        translators = authors.pop("Translators", [])
        version = self.core.call_success("app_get_version")
        icon = self.core.call_success("icon_get_pixbuf", "paperwork", 128)

        about_dialog = widget_tree.get_object("about_dialog")
        about_dialog.set_logo(icon)
        about_dialog.set_version(version)

        for (k, v) in authors.items():
            self._set_authors_list(
                lambda authors: about_dialog.add_credit_section(k, authors),
                v
            )
        if len(translators) > 0:
            self._set_authors_str(
                about_dialog.set_translator_credits, translators
            )
        self._set_authors_list(about_dialog.set_documenters, documentation)

        about_dialog.set_transient_for(self.windows[-1])
        about_dialog.set_visible(True)

        about_dialog.connect("close", self._on_close)
        about_dialog.connect("response", self._on_close)

        self.core.call_all("on_gtk_window_opened", about_dialog)



import os

import openpaperwork_core
import openpaperwork_core.deps


GI_AVAILABLE = False
GLIB_AVAILABLE = False
POPPLER_AVAILABLE = False


try:
    import gi
    GI_AVAILABLE = True
except (ImportError, ValueError):
    pass

if GI_AVAILABLE:
    try:
        from gi.repository import Gio
        GLIB_AVAILABLE = True
    except (ImportError, ValueError):
        pass

    try:
        gi.require_version('Poppler', '0.18')
        from gi.repository import Poppler
        POPPLER_AVAILABLE = True
    except (ImportError, ValueError):
        pass


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'poppler',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GI_AVAILABLE:
            out['gi'] = openpaperwork_core.deps.GI
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)
        if not POPPLER_AVAILABLE:
            out['poppler'] = openpaperwork_core.deps.POPPLER

    def poppler_open(self, url, password=None):
        if os.name == "nt":
            # WORKAROUND(Jflesch):
            # Disabled for now on Windows: There is a file descriptor leak
            # somewhere.
            # While it causes little problems on GNU/Linux, on Windows it
            # prevents deleting documents.
            return None

        gio_file = Gio.File.new_for_uri(url)
        doc = self.core.call_one(
            "mainloop_execute", Poppler.Document.new_from_gfile,
            gio_file, password=password
        )
        self.core.call_all("on_objref_track", doc)
        return doc



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import logging
import shutil

import openpaperwork_core

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            # if there are no doc_renderer loaded, nothing it displayed, which
            # may be fine.
            # (see paperwork-json)
            # {
            #     "interface": "doc_renderer",
            #     "defaults": [],
            # },
            {
                "interface": "document_storage",
                "defaults": ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                "interface": "index",
                "defaults": ['paperwork_backend.index.shell'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'search', help=_("Search keywords in documents")
        )
        p.add_argument(
            '--limit', '-l', type=int, default=50,
            help=_("Maximum number of results (default: 50)")
        )
        p.add_argument(
            'keywords', nargs='*', default=[],
            help=_("Search keywords (none means all documents)")
        )

    def cmd_run(self, console, args):
        if args.command != 'search':
            return None

        keywords = " ".join(args.keywords)

        docs = []
        self.core.call_all("index_search", docs, keywords, args.limit)
        docs.sort(reverse=True)

        renderers = []
        self.core.call_all("doc_renderer_get", renderers)
        renderer = renderers[-1]

        for (doc_id, doc_url) in docs:
            header = _("Document id: %s") % doc_id
            self.core.call_all("print", header)

            doc_date = self.core.call_success("doc_get_date_by_id", doc_id)
            doc_date = self.core.call_success("i18n_date_short", doc_date)
            header = _("Document date: %s") % doc_date
            self.core.call_all("print", header)

            if renderer is None:
                continue
            if doc_url is None:
                LOGGER.warning("Failed to get URL of document %s", doc_id)
                continue
            lines = renderer.get_preview_output(
                doc_id, doc_url, shutil.get_terminal_size()
            )
            for line in lines:
                self.core.call_all("print", line)
            self.core.call_all("print", "")
        self.core.call_success("print_flush")

        return [doc[0] for doc in docs]



"""
Plugin to draw scan on GTK widgets.
This code should be on Paperwork-gtk, but Ironscanner needs it to.
"""
from typing import Any
from typing import List
from typing import Tuple

import logging

import openpaperwork_core
import openpaperwork_core.deps

CAIRO_AVAILABLE = False

try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    pass


LOGGER = logging.getLogger(__name__)


class Drawer(object):
    BACKGROUND = (0.75, 0.75, 0.75)

    def __init__(self, core, drawing_area=None):
        self.core = core
        self.drawing_areas = []
        self.draw_connect_ids = {}

        # expected scan size: note that height may be -1 if unknown
        self.scan_size = (0, 0)

        # Cairo surface size preallocated
        # usually == scan_size unless self.scan_size is weird and smells funny
        self.cairo_size = (0, 0)

        self.last_line = 0
        self.show_scan_border = False
        self.image = None
        self.scan_ended = False

        if drawing_area is not None:
            self.add_drawing_area(drawing_area)

    def add_drawing_area(self, drawing_area):
        self.drawing_areas.append(drawing_area)
        self.draw_connect_ids[drawing_area] = drawing_area.connect(
            "draw", self.on_draw
        )

    def remove_drawing_area(self, drawing_area):
        connect_id = self.draw_connect_ids.pop(drawing_area)
        drawing_area.disconnect(connect_id)
        self.drawing_areas.remove(drawing_area)
        if len(self.drawing_areas) <= 0:
            self.stop()

    def stop(self):
        for (drawing_area, connect_id) in self.draw_connect_ids.items():
            drawing_area.disconnect(connect_id)
        self.draw_connect_ids = {}
        self.drawing_areas = []
        self.image = None  # release the memory

    def request_redraw(self):
        for d in self.drawing_areas:
            d.queue_draw()

    def on_scan_page_start(self, scan_params):
        self.scan_size = (scan_params.get_width(), scan_params.get_height())
        self.last_line = 0
        LOGGER.info(
            "Scan started: %s (expected: %dx%dpx)",
            self.scan_size, self.scan_size[0], self.scan_size[1]
        )
        self.cairo_size = (
            # WORKAROUND(Jflesch): Some scanners (Fujistu mainly) return an
            # image far too big in height. Cairo only allows a maximum image
            # size of 32767 (see #define MAX_IMAGE_SIZE in
            # cairo-image-surface.c)
            min(self.scan_size[0], 32766),
            min(self.scan_size[1], 32766)
            if self.scan_size[1] > 0
            else 32766
        )
        self.image = cairo.ImageSurface(
            cairo.FORMAT_RGB24, self.cairo_size[0], self.cairo_size[1]
        )

        cairo_ctx = cairo.Context(self.image)
        cairo_ctx.set_source_rgb(
            self.BACKGROUND[0], self.BACKGROUND[1], self.BACKGROUND[2]
        )
        cairo_ctx.rectangle(0, 0, self.cairo_size[0], self.cairo_size[1])
        cairo_ctx.fill()

        self.show_scan_border = True
        self.request_redraw()

    def on_scan_page_end(self):
        self.show_scan_border = False
        self.request_redraw()

    def on_scan_chunk(self, img_chunk):
        if self.image is None:
            return

        size = img_chunk.size
        LOGGER.debug("Scan chunk: %s", size)
        img_chunk = self.core.call_success(
            "pillow_to_surface", img_chunk
        )

        cairo_ctx = cairo.Context(self.image)
        cairo_ctx.translate(0, self.last_line)
        cairo_ctx.set_source_surface(img_chunk.surface)
        cairo_ctx.rectangle(0, 0, size[0], size[1])
        cairo_ctx.clip()
        cairo_ctx.paint()

        self.last_line += size[1]
        self.request_redraw()

    def on_draw(self, drawing_area, cairo_ctx):
        widget_height = drawing_area.get_allocated_height()
        widget_width = drawing_area.get_allocated_width()
        factor_w = self.cairo_size[0] / widget_width
        if self.scan_size[1] > 0:
            factor_h = self.cairo_size[1] / widget_height
        else:
            factor_h = self.last_line / widget_height
        factor = max(factor_w, factor_h)

        # background
        cairo_ctx.save()
        try:
            cairo_ctx.set_source_rgb(
                self.BACKGROUND[0], self.BACKGROUND[1], self.BACKGROUND[2]
            )
            cairo_ctx.rectangle(0, 0, widget_width, widget_height)
            cairo_ctx.clip()
            cairo_ctx.paint()
        finally:
            cairo_ctx.restore()

        # chunks
        if self.image is not None:
            cairo_ctx.save()
            try:
                cairo_ctx.scale(1.0 / factor, 1.0 / factor)
                cairo_ctx.set_source_surface(self.image)
                if self.scan_size[1] > 0:
                    cairo_ctx.rectangle(
                        0, 0,
                        self.cairo_size[0], self.cairo_size[1]
                    )
                else:
                    cairo_ctx.rectangle(
                        0, 0,
                        self.cairo_size[0], self.last_line
                    )
                cairo_ctx.clip()
                cairo_ctx.paint()
            finally:
                cairo_ctx.restore()

        # scan border
        if self.show_scan_border:
            cairo_ctx.save()
            try:
                position = int(self.last_line / factor)
                cairo_ctx.set_operator(cairo.OPERATOR_OVER)
                cairo_ctx.set_source_rgba(
                    1.0, 0.0, 0.0, 0.5
                )
                cairo_ctx.set_line_width(10.0)

                cairo_ctx.move_to(0, position)
                cairo_ctx.line_to(widget_width, position)
                cairo_ctx.stroke()
            finally:
                cairo_ctx.restore()


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        # scan id --> Drawer (scan id = None ==> any scan)
        self.active_drawers = {}

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_drawer_scan',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'pillow_to_surface',
                'defaults': ['paperwork_backend.cairo.pillow'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            }
        ]

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'].update(openpaperwork_core.deps.CAIRO)

    def draw_scan_start(self, drawing_area, scan_id=None):
        if scan_id in self.active_drawers:
            drawer = self.active_drawers[scan_id]
        else:
            drawer = Drawer(self.core)
            self.active_drawers[scan_id] = drawer
        drawer.add_drawing_area(drawing_area)
        return drawer

    def draw_scan_stop(self, drawing_area):
        for (k, drawer) in self.active_drawers.items():
            for d in drawer.drawing_areas:
                if d == drawing_area:
                    break
            else:
                continue
            break
        else:
            return None

        drawer.remove_drawing_area(drawing_area)
        if drawer.scan_ended and len(drawer.drawing_areas) <= 0:
            self.active_drawers.pop(k)
        return drawer

    def draw_scan_get_scan_size(self, scan_id):
        active_drawer = self.active_drawers.get(scan_id, None)
        if active_drawer is None:
            return None
        size = active_drawer.cairo_size
        if active_drawer.scan_size[1] < 0:
            size = (size[0], active_drawer.last_line)
        return size

    def on_scan_feed_start(self, scan_id):
        self.core.call_all("on_busy")
        if scan_id not in self.active_drawers:
            return
        # we show the app as busy when the user clicks on 'scan'
        # and we stop the busy indicator when a page is actually scanning
        # instantiate a default drawer to keep track of the size and the chunks
        if scan_id not in self.active_drawers:
            self.active_drawers[scan_id] = Drawer(self.core)

    def on_scan_page_start(self, scan_id, page_nb, scan_params):
        self.core.call_all("on_idle")
        if None in self.active_drawers:
            self.active_drawers[None].on_scan_page_start(scan_params)
        if scan_id in self.active_drawers:
            self.active_drawers[scan_id].on_scan_page_start(scan_params)

    def on_scan_chunk(self, scan_id, scan_params, img_chunk):
        for k in (None, scan_id):
            if k not in self.active_drawers:
                continue
            self.active_drawers[k].on_scan_chunk(img_chunk)

    def on_scan_page_end(self, scan_id, page_nb, img):
        self.core.call_all("on_busy")
        for k in (None, scan_id):
            if k not in self.active_drawers:
                continue
            self.active_drawers[k].on_scan_page_end()
            self.active_drawers[k].scan_ended = True
            if len(self.active_drawers[k].drawing_areas) <= 0:
                self.active_drawers.pop(k)

    def on_scan_feed_end(self, scan_id):
        self.core.call_all("on_idle")
        for k in (None, scan_id):
            if k not in self.active_drawers:
                continue
            self.active_drawers[k].scan_ended = True


if __name__ == "__main__":
    import sys

    import PIL
    import PIL.Image
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import GLib

    img = PIL.Image.open(sys.argv[1])
    chunks = [
        img.crop((0, line, img.size[1], line + 100))
        for line in range(0, img.size[1], 100)
    ]

    class FakeScanParams(object):
        def __init__(self, img):
            self.img = img

        def get_width(self):
            return self.img.size[0]

        def get_height(self):
            return self.img.size[1]

    class FakeModule(object):
        class Plugin(openpaperwork_core.PluginBase):
            def get_interfaces(self):
                return ['scan']

    scan_params = FakeScanParams(img)

    core = openpaperwork_core.Core()
    core.load("openpaperwork_core.mainloop.asyncio")
    core.load("openpaperwork_core.thread.simple")
    core.load("openpaperwork_core.work_queue.default")
    core.load("openpaperwork_gtk.fs.gio")
    core.load("paperwork_backend.cairo.pillow")
    core.load("openpaperwork_core.pillow.img")
    core._load_module("scan", FakeModule)
    core._load_module("test", sys.modules[__name__])
    core.init()

    window = Gtk.Window()
    window.set_size_request(600, 600)

    drawing_area = Gtk.DrawingArea()
    window.add(drawing_area)

    core.call_all("draw_scan_start", drawing_area, scan_id="pouet")

    calls: List[Tuple[str, str, Any, Any]] = [
        ("on_scan_page_start", "pouet", 0, scan_params),
    ]
    calls += [
        ("on_scan_chunk", "pouet", scan_params, chunk)
        for chunk in chunks
    ]
    calls += [
        ("on_scan_page_end", "pouet", 0, img)
    ]

    calls = 2 * calls

    def wrapper(func, *args):
        func(*args)
        return False

    for (t, call) in enumerate(calls):
        t = t * 500 + 500
        GLib.timeout_add(t, wrapper, core.call_all, *call)

    window.show_all()
    Gtk.main()



import math

try:
    import gi
    gi.require_foreign("cairo")
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    CAIRO_AVAILABLE = False

try:
    from gi.repository import GObject
    GI_AVAILABLE = True
except (ImportError, ValueError):
    GI_AVAILABLE = False

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    # workaround so chkdeps can still be called
    class Gtk(object):
        class DrawingArea(object):
            pass
    GTK_AVAILABLE = False


import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps


class LabelWidget(Gtk.DrawingArea):
    FONT = ""
    LABEL_HEIGHT = 23
    LABEL_TEXT_SIZE = 13
    LABEL_TEXT_SHIFT = 3    # Shift a bit to fix valignment
    LABEL_CORNER_RADIUS = 5

    def __init__(self, core, label_txt, label_color, highlight=False):
        super().__init__()
        self.core = core
        self.txt = label_txt
        self.color = label_color
        self.highlight = highlight
        self.connect("draw", self._draw)

        # we must compute the widget size
        dummy = cairo.ImageSurface(cairo.Format.RGB24, 200, 200)
        ctx = cairo.Context(dummy)
        size = self.compute_size(ctx)
        dummy.finish()

        self.set_size_request(size[0], size[1])

    def compute_size(self, cairo_ctx):
        cairo_ctx.set_font_size(self.LABEL_TEXT_SIZE)
        if not self.highlight:
            cairo_ctx.select_font_face(
                self.FONT, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL
            )
        else:
            cairo_ctx.select_font_face(
                self.FONT, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_BOLD
            )

        (p_x1, p_y1, p_x2, p_y2, p_x3, p_y3) = cairo_ctx.text_extents(
            self.txt
        )
        return (
            p_x3 - p_x1 + (2 * self.LABEL_CORNER_RADIUS),
            self.LABEL_HEIGHT
        )

    @staticmethod
    def _rectangle_rounded(cairo_ctx, area, radius):
        (x, y, w, h) = area
        cairo_ctx.new_sub_path()
        cairo_ctx.arc(
            x + w - radius, y + radius,
            radius, -1.0 * math.pi / 2, 0
        )
        cairo_ctx.arc(
            x + w - radius, y + h - radius,
            radius, 0, math.pi / 2
        )
        cairo_ctx.arc(
            x + radius, y + h - radius, radius,
            math.pi / 2, math.pi
        )
        cairo_ctx.arc(
            x + radius, y + radius, radius,
            math.pi, 3.0 * math.pi / 2
        )
        cairo_ctx.close_path()

    def _draw(self, _, cairo_ctx):
        txt_offset = (
            (self.LABEL_HEIGHT - self.LABEL_TEXT_SIZE) / 2 +
            self.LABEL_TEXT_SHIFT
        )

        (w, h) = self.compute_size(cairo_ctx)

        # background rectangle
        bg = self.color
        cairo_ctx.set_source_rgb(bg[0], bg[1], bg[2])
        cairo_ctx.set_line_width(1)
        self._rectangle_rounded(
            cairo_ctx,
            (0, 0, w, h),
            self.LABEL_CORNER_RADIUS
        )
        cairo_ctx.fill()

        # foreground text
        fg = self.core.call_success("label_get_foreground_color", self.color)
        cairo_ctx.set_source_rgb(fg[0], fg[1], fg[2])
        cairo_ctx.move_to(
            self.LABEL_CORNER_RADIUS,
            h - txt_offset
        )
        cairo_ctx.text_path(self.txt)
        cairo_ctx.fill()


if GTK_AVAILABLE:
    GObject.type_register(LabelWidget)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_widget_label',
        ]

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'] = openpaperwork_core.deps.CAIRO
        if not GI_AVAILABLE:
            out['gi'] = openpaperwork_core.deps.GI
        if not GTK_AVAILABLE:
            out['gtk'] = openpaperwork_gtk.deps.GTK

    def gtk_widget_label_new(self, label_txt, label_color, highlight=False):
        assert CAIRO_AVAILABLE
        assert GI_AVAILABLE
        assert GTK_AVAILABLE
        return LabelWidget(self.core, label_txt, label_color, highlight)



"""
This plugin is an helper for other plugins. It provides an easy way to
detect deleted, modified or added documents when synchronizing with the work
directory.
"""

import datetime
import logging

import openpaperwork_core

from . import (_, sync)

LOGGER = logging.getLogger(__name__)

CREATE_TABLES = [
    (
        "CREATE TABLE IF NOT EXISTS documents ("
        " doc_id TEXT PRIMARY KEY,"
        " text TEXT NULL,"
        " mtime INTEGER NOT NULL"
        ")"
    ),
]

ID = "doctracker"


class DocTrackerTransaction(sync.BaseTransaction):
    def __init__(self, plugin, total_expected=-1):
        super().__init__(plugin.core, total_expected)
        self.priority = -10000

        # This transaction uses its own SqlLite connection to avoid issues
        # when the user switches from one work directory to another while
        # a transaction is still running.
        self.sql = self.core.call_one(
            "sqlite_execute",
            self.core.call_success,
            "sqlite_open", plugin.sql_file
        )

        self.core.call_one(
            "sqlite_execute", self.sql.execute, "BEGIN TRANSACTION"
        )

    def _get_actual_doc_data(self, doc_id, doc_url):
        if (
                    doc_url is not None
                    and self.core.call_success("fs_exists", doc_url)
                    is not None
                ):
            mtime = self.core.call_success("doc_get_mtime_by_url", doc_url)
            if mtime is None:
                mtime = 0

            doc_text = []
            self.core.call_all("doc_get_text_by_url", doc_text, doc_url)
            doc_text = "\n\n".join(doc_text)
        else:
            mtime = 0
            doc_text = None

        return {
            'mtime': mtime,
            'text': doc_text,
        }

    def add_doc(self, doc_id):
        self.notify_progress(ID, _("Document %s added") % (doc_id))
        self._upd_doc(doc_id)
        super().add_doc(doc_id)

    def upd_doc(self, doc_id):
        self.notify_progress(ID, _("Document %s updated") % (doc_id))
        self._upd_doc(doc_id)
        super().upd_doc(doc_id)

    def _upd_doc(self, doc_id):
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        actual = self._get_actual_doc_data(doc_id, doc_url)

        self.core.call_one(
            "sqlite_execute", self.sql.execute,
            "INSERT OR REPLACE INTO documents (doc_id, text, mtime)"
            " VALUES (?, ?, ?)",
            (doc_id, actual['text'], actual['mtime'])
        )

    def del_doc(self, doc_id):
        self.notify_progress(ID, _("Document %s deleted") % (doc_id))
        self.core.call_one(
            "sqlite_execute", self.sql.execute,
            "DELETE FROM documents WHERE doc_id = ?",
            (doc_id,)
        )
        super().del_doc(doc_id)

    def unchanged_doc(self, doc_id):
        self.notify_progress(
            ID, _("Examining document %s: unchanged") % (doc_id)
        )
        super().unchanged_doc(doc_id)

    def cancel(self):
        self.notify_progress(ID, _("Rolling back changes"))
        self.core.call_one("sqlite_execute", self.sql.execute, "ROLLBACK")
        self.core.call_one("sqlite_execute", self.sql.close)
        self.notify_done(ID)

    def commit(self):
        self.notify_progress(ID, _("Committing changes"))
        self.core.call_one("sqlite_execute", self.sql.execute, "COMMIT")
        self.core.call_one("sqlite_execute", self.sql.close)
        self.notify_done(ID)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def __init__(self):
        self.sql = None
        self.transaction_factories = []

    def get_interfaces(self):
        return [
            'doc_tracking',
            'syncable',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio']
            },
            {
                'interface': 'data_dir_handler',
                'defaults': ['paperwork_backend.datadirhandler'],
            },
            {
                'interface': 'sqlite',
                'defaults': ['openpaperwork_core.sqlite'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def init(self, core):
        super().init(core)
        self._init()

    def _init(self):
        paperwork_dir = self.core.call_success(
            "data_dir_handler_get_individual_data_dir"
        )
        self.sql_file = self.core.call_success(
            "fs_join", paperwork_dir, 'doc_tracking.db'
        )
        self.sql = self.core.call_one(
            "sqlite_execute",
            self.core.call_success,
            "sqlite_open", self.sql_file
        )
        for query in CREATE_TABLES:
            self.core.call_one("sqlite_execute", self.sql.execute, query)

    def on_quit(self):
        LOGGER.info("Closing doc tracker db ...")
        self.core.call_one(
            "sqlite_execute",
            self.core.call_success,
            "sqlite_close", self.sql
        )
        self.sql = None

    def on_data_dir_changed(self):
        self.on_quit()
        self._init()

    def doc_tracker_register(self, name, transaction_factory):
        self.transaction_factories.append((name, transaction_factory))

    def doc_tracker_get_all_doc_ids(self):
        doc_ids = self.core.call_one(
            "sqlite_execute", self.sql.execute,
            "SELECT doc_id FROM documents"
        )
        doc_ids = self.core.call_one(
            "sqlite_execute", list, doc_ids
        )
        return {doc_id[0] for doc_id in doc_ids}

    def doc_tracker_get_doc_text_by_id(self, doc_id):
        """
        Return the text of a document as last-known.
        This method is useful when a document is deleted: when a plugin
        is notified that a document has been deleted, it may still need its
        text. Since the document won't be available anymore, we pull its text
        from the database (for instance, for label guesser untraining)
        """
        text = self.core.call_one(
            "sqlite_execute", self.sql.execute,
            "SELECT text FROM documents WHERE doc_id = ? LIMIT 1",
            (doc_id,)
        )
        text = self.core.call_one(
            "sqlite_execute", list, text
        )
        if len(text) <= 0:
            return None
        return text[0][0]

    def doc_transaction_start(self, out: list, total_expected=-1):
        for (name, transaction_factory) in self.transaction_factories:
            out.append(transaction_factory(
                sync=False, total_expected=total_expected
            ))
        out.append(DocTrackerTransaction(self, total_expected))

    def sync(self, promises: list):
        storage_all_docs = []
        names = [t[0] for t in self.transaction_factories]
        names.append('doc_tracker')

        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self.core.call_all,
            args=("storage_get_all_docs", storage_all_docs,)
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(storage_all_docs.sort)

        class DbDoc(object):
            def __init__(self, result):
                self.key = result[0]
                self.extra = datetime.datetime.fromtimestamp(result[1])

        promise = promise.then(self.sql.cursor)
        promise = promise.then(lambda cursor: (
            [
                sync.StorageDoc(self.core, doc[0], doc[1])
                for doc in storage_all_docs
            ],
            [
                DbDoc(r)
                for r in cursor.execute("SELECT doc_id, mtime FROM documents")
            ],
        ))
        promise = promise.then(lambda args: (
            *args,
            sorted([
                transaction_factory(
                    sync=True,
                    total_expected=max(len(storage_all_docs), len(args[1]))
                )
                for (name, transaction_factory) in self.transaction_factories
            ] + [
                DocTrackerTransaction(
                    self,
                    total_expected=max(len(storage_all_docs), len(args[1]))
                )
            ], key=lambda t: -1 * t.priority),
        ))
        promise = promise.then(lambda args: sync.Syncer(
            self.core, names, args[0], args[1], args[2]
        ))
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, lambda syncer: syncer.run()
        ))
        promises.append(promise)

    def tests_cleanup(self):
        self.sql.close()



import gc
import logging
import tempfile
import weakref

import objgraph

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.objs = []
        self.graph_path = tempfile.mktemp(suffix='.png')

    def get_interfaces(self):
        return ['memleak_detector']

    def on_objref_track(self, obj):
        assert obj is not None
        self.objs.append((str(type(obj)), weakref.ref(obj)))

    def on_objref_graph(self):
        to_graph = []

        objs = self.objs.copy()
        gc.collect()
        for (idx, (type_name, wref)) in reversed(list(enumerate(objs))):
            obj = wref()
            if obj is None:
                LOGGER.info("Object of type %s has disappeared", type_name)
                self.objs.pop(idx)
                continue
            to_graph.append(obj)

        if len(to_graph) <= 0:
            LOGGER.info("Nothing to graph")
            return

        LOGGER.info(
            "Making reference graph for %d objects to %s",
            len(to_graph), self.graph_path
        )
        objgraph.show_backrefs(to_graph, filename=self.graph_path)

    def on_memleak_track_stop(self):
        LOGGER.info("Most common object types:")
        objgraph.show_most_common_types()



from .. import PluginBase


class Setting(object):
    def __init__(self, value, default_func):
        self.value = value
        self.default_value_func = default_func
        self.observers = []

    def get(self):
        if self.value is None:
            return self.default_value_func()
        return self.value

    def put(self, v):
        self.value = v
        for obs in self.observers:
            obs()

    def add_observer(self, obs):
        self.observers.append(obs)

    def remove_observer(self, obs):
        self.observers.remove(obs)


class Plugin(PluginBase):
    """
    Translate values from the configuration into more usable ones.
    Provides default values (except for plugins).
    """

    def __init__(self):
        super().__init__()
        self._settings = {}

    def __get_settings(self):
        return {k: v.get() for (k, v) in self._settings.items()}

    def __set_settings(self, new_settings):
        self._settings = {
            k: Setting(v, lambda: None)
            for (k, v) in new_settings.items()
        }

    settings = property(__get_settings, __set_settings)

    def get_interfaces(self):
        return ['config']

    def config_load(self):
        pass

    def config_load_plugins(self, plugin_list_name, default_plugins=[]):
        pass

    def config_save(self):
        pass

    def config_build_simple(self, section, token, default):
        return Setting(None, default)

    def config_register(self, key, setting):
        if key not in self._settings:  # don't smash test settings
            self._settings[key] = setting

    def config_get_setting(self, key):
        return self._settings[key]

    def config_get(self, key):
        return self._settings[key].get()

    def config_get_default(self, key):
        return self._settings[key].default_value_func()

    def config_put(self, key, value):
        self._settings[key].put(value)

    def config_add_plugin(self, plugin):
        raise NotImplementedError()

    def config_remove_plugin(self, plugin):
        raise NotImplementedError()

    def config_list_plugins(self):
        raise NotImplementedError()

    def config_add_observer(self, key: str, callback):
        self._settings[key].add_observer(callback)

    def config_remove_observer(self, key: str, callback):
        self._settings[key].remove_observer(callback)



"""
Used to archive logs, screenshots, etc.
Do not depends on 'fs' plugins (this plugin is used for logging and therefore
is critical --> has minimum dependencies).
"""

import datetime
import logging

from . import PluginBase


LOGGER = logging.getLogger(__name__)
ARCHIVE_FILE_DATE_FORMAT = "%Y%m%d_%H%M_%S"
MAX_DAYS = 31


class ArchiveHandler(object):
    def __init__(self, core, storage_name, storage_dir, file_extension):
        self.core = core
        self.storage_name = storage_name
        self.storage_dir = storage_dir
        self.file_extension = file_extension

    def get_new(self, name=None):
        if name is None:
            name = self.storage_name
        out_file_name = datetime.datetime.now().strftime(
            ARCHIVE_FILE_DATE_FORMAT
        )
        out_file_name += "_{}.{}".format(name, self.file_extension)
        out_file_path = self.core.call_success(
            "fs_join", self.storage_dir, out_file_name
        )
        return out_file_path

    def get_archived(self):
        for f in self.core.call_success("fs_listdir", self.storage_dir):
            f = self.core.call_success("fs_basename", f)
            if not f.lower().endswith(".{}".format(self.file_extension)):
                continue
            short_f = "_".join(f.split("_", 3)[:3])
            try:
                date = datetime.datetime.strptime(
                    short_f, ARCHIVE_FILE_DATE_FORMAT
                )
            except ValueError as exc:
                LOGGER.warning(
                    "Unexpected filename: %s. Ignoring it",
                    f, exc_info=exc
                )
                continue
            yield (
                date, self.core.call_success("fs_join", self.storage_dir, f)
            )

    def delete_obsoletes(self):
        now = datetime.datetime.now()
        for (date, file_path) in self.get_archived():
            if (now - date).days <= MAX_DAYS:
                continue
            LOGGER.info("Deleting obsolete log file: %s", file_path)
            self.core.call_success("fs_unlink", file_path, trash=False)


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.storage_dirs = []

    def get_interfaces(self):
        return ['file_archives']

    def get_deps(self):
        return [
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'paths',
                'defaults': ['openpaperwork_core.paths.xdg'],
            },
        ]

    def init(self, core):
        super().init(core)
        data_dir = self.core.call_success("paths_get_data_dir")
        self.base_archive_dir = self.core.call_success(
            "fs_join", data_dir, "openpaperwork"
        )

        LOGGER.info("Archiving to %s", self.base_archive_dir)

    def file_archive_get(self, storage_name, file_extension):
        self.core.call_success("fs_mkdir_p", self.base_archive_dir)
        storage_dir = self.core.call_success(
            "fs_join", self.base_archive_dir, storage_name
        )
        self.storage_dirs.append(storage_dir)
        LOGGER.info("Archiving '%s' to %s", storage_name, storage_dir)
        self.core.call_success("fs_mkdir_p", storage_dir)
        archiver = ArchiveHandler(
            self.core, storage_name, storage_dir, file_extension
        )
        archiver.delete_obsoletes()
        return archiver



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.windows = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_dialog_single_entry',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def gtk_show_dialog_single_entry(
            self, origin, title, original_value, *args, **kwargs):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.dialogs.single_entry", "single_entry.glade"
        )
        widget_tree.get_object("entry").set_text(original_value)

        dialog = widget_tree.get_object("dialog")
        dialog.set_title(title)
        dialog.set_transient_for(self.windows[-1])
        dialog.set_modal(True)
        dialog.connect(
            "response", self._on_response, (widget_tree, origin, args, kwargs)
        )
        dialog.show_all()
        return True

    def _on_response(self, dialog, response_id, args):
        (widget_tree, origin, args, kwargs) = args
        new_value = widget_tree.get_object("entry").get_text()
        dialog.destroy()

        if (response_id != 0 and
                response_id != Gtk.ResponseType.ACCEPT and
                response_id != Gtk.ResponseType.OK and
                response_id != Gtk.ResponseType.YES and
                response_id != Gtk.ResponseType.APPLY):
            LOGGER.info("User cancelled")
            r = False
        else:
            r = True
        self.core.call_all(
            "on_dialog_single_entry_reply",
            origin, r, new_value,
            *args, **kwargs
        )



import unittest

from .. import Core


class AbstractTest(unittest.TestCase):
    def get_plugin_name(self):
        """
        Must be overloaded by subclasses
        """
        assert False

    def setUp(self):
        self.core = Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_basic(self):
        boxes = [
            (((0, 0), (10, 10)), "a"),
            (((10, 10), (50, 40)), "b"),
            (((25, 25), (100, 100)), "c"),
        ]
        indexer = self.core.call_success("spatial_indexer_get", boxes)
        self.assertEqual(
            list(indexer.get_boxes(5, 5)),
            [
                (((0, 0), (10, 10)), "a"),
            ]
        )
        self.assertEqual(
            list(indexer.get_boxes(5, 10)),
            [
                (((0, 0), (10, 10)), "a"),
            ]
        )
        self.assertEqual(
            list(indexer.get_boxes(10, 10)),
            [
                (((0, 0), (10, 10)), "a"),
                (((10, 10), (50, 40)), "b"),
            ]
        )
        self.assertEqual(
            list(indexer.get_boxes(25, 20)),
            [
                (((10, 10), (50, 40)), "b"),
            ]
        )
        self.assertEqual(
            list(indexer.get_boxes(75, 75)),
            [
                (((25, 25), (100, 100)), "c"),
            ]
        )

    def test_real(self):
        boxes = [
            (((584, 360), (1097, 415)), "CROIX-ROUGE"),
            (((1126, 361), (1540, 428)), "FRANCAISE"),
            (((865, 430), (1006, 464)), "EQUIPES"),
            (((1028, 437), (1258, 462)), "SECOURISTES"),
            (((724, 675), (911, 718)), "Madame"),
        ]
        indexer = self.core.call_success("spatial_indexer_get", boxes)
        self.assertEqual(
            list(indexer.get_boxes(836, 701)),
            [
                (((724, 675), (911, 718)), "Madame"),
            ]
        )



"""
Rendering cache: actually used to keep a copy of the rendered page at the
expected size, and only recompute it when the zoom level changes.
Otherwise we overuse Poppler and some PDF may be slow to render correctly
(those with full-page images in it for instance)
"""

import logging

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise

CAIRO_AVAILABLE = False
GLIB_AVAILABLE = False

try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    pass

try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    # dummy so chkdeps can still be called
    class GObject(object):
        class SignalFlags(object):
            RUN_LAST = 0

        class GObject(object):
            pass


LOGGER = logging.getLogger(__name__)


class CacheRenderer(GObject.GObject):
    __gsignals__ = {
        'getting_size': (GObject.SignalFlags.RUN_LAST, None, ()),
        'size_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'img_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
    }
    OUTLINE = (0.5, 0.5, 0.5)

    def __init__(
            self, core, work_queue_name, wrapped_renderer):
        super().__init__()
        self.core = core
        self.work_queue_name = work_queue_name
        self.wrapped_renderer = wrapped_renderer

        wrapped_renderer.connect(
            "getting_size", self._on_wrapped_getting_size
        )
        wrapped_renderer.connect(
            "size_obtained", self._on_wrapped_size_obtained
        )
        wrapped_renderer.connect(
            "img_obtained", self._on_wrapped_img_obtained
        )

        self.rendering = None

    def __str__(self):
        return f"CacheRenderer({self.wrapped_renderer})"

    def _on_wrapped_getting_size(self, wrapped_renderer):
        self.emit('getting_size')

    def _on_wrapped_size_obtained(self, wrapped_renderer):
        self.emit('size_obtained')

    def _on_wrapped_img_obtained(self, wrapped_renderer):
        self.rendering = None
        self.emit('img_obtained')

    @property
    def size(self):
        return self.wrapped_renderer.size

    def _set_zoom(self, z):
        self.rendering = None
        self.wrapped_renderer.zoom = z

    def _get_zoom(self):
        return self.wrapped_renderer.zoom

    zoom = property(_get_zoom, _set_zoom)

    def start(self):
        self.wrapped_renderer.start()

    def render(self):
        self.wrapped_renderer.render()

    def hide(self):
        self.wrapped_renderer.hide()

    def close(self):
        self.rendering = None
        self.wrapped_renderer.close()

    def draw(self, cairo_ctx):
        if self.rendering is None:
            LOGGER.info("Cache miss. Rendering")
            renderer = self.wrapped_renderer
            rendering_w = int(renderer.size[0] * renderer.zoom)
            rendering_h = int(renderer.size[1] * renderer.zoom)
            rendering = cairo.ImageSurface(
                cairo.FORMAT_ARGB32,
                rendering_w, rendering_h
            )
            rendering_ctx = cairo.Context(rendering)
            renderer.draw(rendering_ctx)
            self.rendering = rendering

        cairo_ctx.save()
        try:
            cairo_ctx.set_source_surface(self.rendering)
            cairo_ctx.paint()
        finally:
            cairo_ctx.restore()

    def blur(self):
        self.rendering = None
        self.wrapped_renderer.blur()

    def unblur(self):
        self.rendering = None
        self.wrapped_renderer.unblur()


if GLIB_AVAILABLE:
    GObject.type_register(CacheRenderer)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def get_interfaces(self):
        return [
            'cairo_url',
            'chkdeps',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'urls',
                'defaults': ['openpaperwork_core.urls'],
            },
        ]

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'].update(openpaperwork_core.deps.CAIRO)
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def cairo_renderer_by_url(
            self, work_queue_name, file_url,
            cache=True, **kwargs):
        if not cache:
            return None

        wrapped_renderer = self.core.call_success(
            "cairo_renderer_by_url",
            work_queue_name, file_url, cache=False, **kwargs
        )
        if wrapped_renderer is None:
            LOGGER.error("Unable to get renderer for '%s' !?", file_url)
            return None

        return CacheRenderer(
            self.core, work_queue_name, wrapped_renderer
        )



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import logging
import shutil

import rich.text

import openpaperwork_core

from . import util
from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            # optional dependency
            # {
            #     "interface": "img_renderer",
            #     "defaults": ["paperwork_shell.display.docrendering.img"],
            # },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                "interface": "page_reset",
                "defaults": ["paperwork_backend.model.img_overlay"],
            },
            {
                "interface": "pillow",
                "defaults": [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        reset_parser = parser.add_parser(
            'reset', help=_("Reset a page to its original content")
        )
        # for safety, we mark the argument --pages as required
        reset_parser.add_argument('--pages', '-p', type=str, required=True)
        reset_parser.add_argument('doc_id')

    def show_page(self, console, doc_url, page_idx):
        page_url = self.core.call_success(
            "page_get_img_url", doc_url, page_idx
        )
        img = self.core.call_success("url_to_pillow", page_url)

        terminal_width = shutil.get_terminal_size()[0] - 1
        img = self.core.call_success(
            "img_render", img, terminal_width=terminal_width
        )
        if img is None:
            return
        for line in img:
            console.print(rich.text.Text(line))
        console.print("")

    def cmd_run(self, console, args):
        if args.command != 'reset':
            return None

        doc_id = args.doc_id
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        pages = util.parse_page_list(args)
        out = []

        for page_idx in pages:
            out.append((doc_id, page_idx))
            console.print(
                _("Reseting document {} page {} ...").format(
                    doc_id, page_idx
                )
            )
            console.print(_("Original:"))
            self.show_page(doc_url, page_idx)

            self.core.call_all("page_reset_by_url", doc_url, page_idx)

            console.print(_("Reseted:"))
            self.show_page(doc_url, page_idx)
            console.print("")

        self.core.call_success("transaction_simple", [("upd", doc_id),])
        self.core.call_success("mainloop_quit_graceful")
        self.core.call_success("mainloop")

        console.print(_("Done"))
        console.print(_("All done !"))

        return out



import logging

import openpaperwork_core

from . import BaseDocViewController


LOGGER = logging.getLogger(__name__)


class PageNumberController(BaseDocViewController):
    def _get_closest_widget(self):
        vadj = self.plugin.scroll.get_vadjustment()
        view_width = self.plugin.scroll.get_allocated_width()
        view_height = self.plugin.scroll.get_allocated_height()
        center = (
            view_width / 2,
            vadj.get_value() + (view_height / 2)
        )

        # look if the center is precisely on in a widget
        for widget in self.plugin.page_layout.get_children():
            if not widget.get_visible():
                continue
            alloc = widget.get_allocation()
            if alloc.width < 10 or alloc.height < 10:
                continue
            if (alloc.x <= center[0] and alloc.y <= center[1] and
                    alloc.x + alloc.width >= center[0] and
                    alloc.y + alloc.height >= center[1]):
                return widget

        # else look for the closest widget
        min_dist = (99999999999999999, None)
        for widget in self.plugin.page_layout.get_children():
            if not widget.get_visible():
                continue
            alloc = widget.get_allocation()
            if alloc.width < 10 or alloc.height < 10:
                continue
            widget_center = (
                (alloc.x + (alloc.width / 2)),
                (alloc.y + (alloc.height / 2)),
            )
            dist_w = (center[0] - widget_center[0])
            dist_h = (center[1] - widget_center[1])
            dist = (dist_w * dist_w) + (dist_h * dist_h)
            if dist < min_dist[0]:
                min_dist = (dist, widget)
        return min_dist[1]

    def _update_current_page(self):
        widget = self._get_closest_widget()
        if widget is None:
            return
        page = self.plugin.widget_to_page[widget]
        self.plugin.core.call_all("on_page_shown", page.page_idx)

    def on_layout_size_allocate(self, layout):
        super().on_layout_size_allocate(layout)
        self._update_current_page()

    def on_page_size_obtained(self, layout):
        super().on_page_size_obtained(layout)
        self._update_current_page()

    def on_vscroll_changed(self, vadj):
        super().on_vscroll_changed(vadj)
        self._update_current_page()

    def on_vscroll_value_changed(self, vadj):
        super().on_vscroll_value_changed(vadj)
        self._update_current_page()


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['gtk_docview_controller']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
        ]

    def gtk_docview_get_controllers(self, out: dict, docview):
        out['page_number'] = PageNumberController(docview)



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class DialogBuilder(object):
    def __init__(self, plugin, title):
        self.plugin = plugin
        self.widget_tree = self.plugin.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.notifications",
            "dialog.glade"
        )
        dialog = self.widget_tree.get_object("dialog")
        dialog.set_title(title)
        dialog.connect("response", self._on_response)

        label = self.widget_tree.get_object("message")
        label.set_text(title)
        label.set_visible(True)

    def _on_response(self, dialog, response):
        dialog.destroy()

    def set_message(self, message):
        label = self.widget_tree.get_object("message")
        label.set_text(message)
        return self

    def set_icon(self, icon):
        self.widget_tree.get_object("dialog").set_icon_name(icon)
        return self

    def add_action(self, action_id, label, callback, *args, **kwargs):
        buttons = self.widget_tree.get_object("buttons")
        button = Gtk.Button.new_with_label(label)
        button.connect("clicked", self._on_click, callback, args, kwargs)
        button.set_visible(True)
        buttons.add(button)
        return self

    def _on_click(self, button, callback, args, kwargs):
        self.widget_tree.get_object("dialog").destroy()
        callback(*args, **kwargs)

    def set_image_from_pixbuf(self, pixbuf):
        img = self.widget_tree.get_object("image")
        img.set_from_pixbuf(pixbuf)
        img.set_visible(True)
        return self

    def show(self):
        dialog = self.widget_tree.get_object("dialog")
        dialog.set_transient_for(self.plugin.windows[-1])
        dialog.set_visible(True)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.windows = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_window_listener',
            'notifications',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def get_notification_builder(self, title, need_actions=False):
        return DialogBuilder(self, title)



from typing import List
from typing import Tuple
import logging
import re

import openpaperwork_core
import paperwork_backend.model.pdf

from .. import _


LOGGER = logging.getLogger(__name__)

PAGE_FILENAME_REGEX = re.compile(r"paper\.(\d+)\.(.*)")
PAGE_FILENAME_FMT = "paper.{idx}.{file_ext}"


class Plugin(openpaperwork_core.PluginBase):
    """
    Label files in each document contains both the label names and their color.
    If the user changes a label color, label files are updated one-by-one.
    If this process is interrupted, we end up with 2 colors for the same
    label.
    """

    PRIORITY = 10000

    def get_interfaces(self):
        return ['chkworkdir']

    def get_deps(self):
        return [
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'doc_pdf_url',
                'defaults': ['paperwork_backend.model.pdf'],
            },
        ]

    def check_work_dir(self, out_problems: list):
        all_docs: List[Tuple[str, str]] = []
        self.core.call_all("storage_get_all_docs", all_docs, only_valid=False)
        all_docs.sort()

        total = len(all_docs)
        LOGGER.info("Checking work directory (%d documents)", total)

        for (idx, (doc_id, doc_url)) in enumerate(all_docs):
            self.core.call_all(
                "on_progress", "chkworkdir_corrupted_page_map",
                idx / total, _("Checking doc %s") % (doc_id,)
            )

            page_map_url = self.core.call_success(
                "fs_join", doc_url,
                paperwork_backend.model.pdf.PdfPageMapping.MAPPING_FILE
            )
            if not self.core.call_success("fs_exists", page_map_url):
                continue

            # simplest way to see if the page map is readable is to
            # try to get the number of pages of the doc
            try:
                self.core.call_success("doc_get_nb_pages_by_url", doc_url)
                continue
            except Exception:
                out_problems.append({
                    "problem": "corrupted_page_map",
                    "doc_id": doc_id,
                    "doc_url": doc_url,
                    "human_description": {
                        "problem": (
                            _(
                                "Page mapping of document %s has been"
                                " corrupted"
                            ) % (doc_id,)
                        ),
                        "solution": (
                            _(
                                "Page mapping of document %s must be"
                                " reinitialized"
                            ) % (doc_id,)
                        )
                    }
                })

        self.core.call_all("on_progress", "chkworkdir_corrupted_page_map", 1.0)

    def fix_work_dir(self, problems):
        total = len(problems)
        for (idx, problem) in enumerate(problems):
            if problem['problem'] != 'corrupted_page_map':
                continue
            doc_id = problem['doc_id']
            doc_url = problem['doc_url']
            LOGGER.info("Fixing document %s", doc_url)
            self.core.call_all(
                "on_progress", "fixworkdir_corrupted_page_map",
                idx / total,
                _("Fixing page mapping of doc %s") % (doc_id,)
            )
            page_map_url = self.core.call_success(
                "fs_join", doc_url,
                paperwork_backend.model.pdf.PdfPageMapping.MAPPING_FILE
            )
            if self.core.call_success("fs_exists", page_map_url):
                self.core.call_success("fs_unlink", page_map_url)

            # Move all the scanned page files after the PDF pages
            doc_nb_pages = self.core.call_success(
                "doc_pdf_get_real_nb_pages_by_url", doc_url
            )
            page_map = {}
            files = self.core.call_success("fs_listdir", doc_url)
            if files is None:
                return None
            for file_url in files:
                file_name = self.core.call_success("fs_basename", file_url)
                match = PAGE_FILENAME_REGEX.match(file_name)
                if match is None:
                    continue
                old_page_nb = int(match.group(1))
                file_ext = match.group(2)
                if old_page_nb in page_map:
                    new_page_nb = page_map[old_page_nb]
                else:
                    new_page_nb = doc_nb_pages + 1
                    doc_nb_pages += 1
                    page_map[old_page_nb] = new_page_nb
                new_page_url = self.core.call_success(
                    "fs_join", doc_url, PAGE_FILENAME_FMT.format(
                        idx=new_page_nb,
                        file_ext=file_ext
                    )
                )
                self.core.call_success("fs_rename", file_url, new_page_url)

        self.core.call_all("on_progress", "fixworkdir_corrupted_page_map", 1.0)



"""
This plugin is an helper for other plugins. It provides an easy way
to track pages that have not yet being treated in each document.

For instance, when a document is notified as updated (see transactions), the
OCR plugin needs to know which pages of this document have already been
OCR-ed and which haven't.
"""
import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)

CREATE_TABLES = [
    (
        "CREATE TABLE IF NOT EXISTS pages ("
        " doc_id TEXT NOT NULL,"
        " page INTEGER NOT NULL,"
        " hash TEXT NOT NULL,"
        " PRIMARY KEY (doc_id, page)"
        ")"
    ),
]


class PageTracker(object):
    def __init__(self, core, sql_url):
        self.core = core
        self.sql = self.core.call_one(
            "sqlite_execute",
            self.core.call_success,
            "sqlite_open", sql_url
        )
        for query in CREATE_TABLES:
            self.core.call_one("sqlite_execute", self.sql.execute, query)

        self.core.call_one(
            "sqlite_execute", self.sql.execute, "BEGIN TRANSACTION"
        )

    def _close(self):
        LOGGER.info("Closing page tracker db ...")
        self.core.call_one(
            "sqlite_execute",
            self.core.call_success,
            "sqlite_close", self.sql
        )
        self.sql = None

    def cancel(self):
        if self.sql is None:
            return
        self.core.call_one(
            "sqlite_execute", self.sql.execute, "ROLLBACK"
        )
        self._close()

    def commit(self):
        if self.sql is None:
            return
        self.core.call_one("sqlite_execute", self.sql.execute, "COMMIT")
        self._close()

    def find_changes(self, doc_id, doc_url):
        """
        Examine a document. Return page that haven't been handled yet
        or that have been modified since.
        Don't forget to call ack_page() once you've handled each page.
        """

        out = []

        db_pages = self.core.call_one(
            "sqlite_execute", self.sql.execute,
            "SELECT page, hash FROM pages"
            " WHERE doc_id = ?",
            (doc_id,)
        )
        db_pages = self.core.call_one(
            "sqlite_execute",
            lambda pages: {r[0]: int(r[1], 16) for r in pages}, db_pages
        )
        db_hashes = set(db_pages.values())

        fs_nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", doc_url
        ) or 0
        fs_pages = {}
        for page_idx in range(0, fs_nb_pages):
            fs_pages[page_idx] = self.core.call_success(
                "page_get_hash_by_url", doc_url, page_idx
            )

        for (page_idx, fs_page_hash) in fs_pages.items():
            if page_idx not in db_pages:
                out.append(('new', page_idx))
            else:
                db_page_hash = db_pages.pop(page_idx)
                if db_page_hash != fs_page_hash:
                    if fs_page_hash in db_hashes:
                        # this page existed before
                        out.append(('moved', page_idx))
                    else:
                        out.append(('upd', page_idx))

        for (db_page_idx, h) in db_pages.items():
            self.core.call_one(
                "sqlite_execute", self.sql.execute,
                "DELETE FROM pages"
                " WHERE doc_id = ? AND page = ?",
                (doc_id, db_page_idx)
            )

        return out

    def ack_page(self, doc_id, doc_url, page_idx):
        """
        Mark the page update has handled.
        """
        page_hash = self.core.call_success(
            "page_get_hash_by_url", doc_url, page_idx
        )
        self.core.call_one(
            "sqlite_execute", self.sql.execute,
            "INSERT OR REPLACE"
            " INTO pages (doc_id, page, hash)"
            " VALUES (?, ?, ?)",
            (doc_id, page_idx, format(page_hash, 'x'))
        )

    def delete_doc(self, doc_id):
        self.core.call_one(
            "sqlite_execute", self.sql.execute,
            "DELETE FROM pages WHERE doc_id = ?",
            (doc_id,)
        )


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        pass

    def get_interfaces(self):
        return ['page_tracking']

    def get_deps(self):
        return [
            {
                'interface': 'data_dir_handler',
                'defaults': ['paperwork_backend.datadirhandler'],
            },
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio']
            },
            {
                'interface': 'page_boxes',
                'defaults': ['paperwork_backend.model.hocr'],
            },
            {
                'interface': 'page_img',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'sqlite',
                'defaults': ['openpaperwork_core.sqlite'],
            },
        ]

    def page_tracker_get(self, tracking_id):
        paperwork_dir = self.core.call_success(
            "data_dir_handler_get_individual_data_dir"
        )
        sql_file = self.core.call_success(
            "fs_join", paperwork_dir,
            'page_tracking_{}.db'.format(tracking_id)
        )
        return PageTracker(self.core, sql_file)



import logging

import pillowfight

import openpaperwork_core
import openpaperwork_core.promise

import paperwork_backend.cairo.pillow


from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.settings_widget_tree = None
        self.size_allocate_connect_id = None
        self.scan_height = 0
        self.scan_width = 0
        self.scan_img = None

    def get_interfaces(self):
        return [
            'gtk_settings_calibration',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_drawer_calibration',
                'defaults': ['paperwork_gtk.drawer.calibration'],
            },
            {
                'interface': 'gtk_drawer_pillow',
                'defaults': ['openpaperwork_gtk.drawer.pillow'],
            },
            {
                'interface': 'gtk_drawer_scan',
                'defaults': [
                    'openpaperwork_gtk.drawer.scan',
                    'paperwork_gtk.drawer.calibration',
                ],
            },
            # Optional:
            # {
            #     'interface': 'gtk_zoomable',
            #     'defaults': [
            #         'paperwork_gtk.gesture.zoom',
            #         'paperwork_gtk.keyboard_shortcut.zoom',
            #     ],
            # },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_settings_dialog',
                'defaults': ['paperwork_gtk.settings'],
            },
            {
                'interface': 'gtk_settings_scanner',
                'defaults': ['paperwork_gtk.settings.scanner.settings'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def complete_settings(self, settings_widget_tree):
        self.settings_widget_tree = settings_widget_tree
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.settings.scanner", "calibration.glade"
        )

        self.size_allocate_connect_id = self.widget_tree.get_object(
            "calibration_scroll"
        ).connect(
            "size-allocate", self._update_calibration_scroll_area_size
        )
        self.value_changed_connect_id = self.widget_tree.get_object(
            "calibration_scale_adjustment"
        ).connect(
            "value-changed", self._update_calibration_area_size_based_on_scale
        )

        self.widget_tree.get_object("calibration_back").connect(
            "clicked", self.hide_calibration_screen
        )
        self.widget_tree.get_object("calibration_scan").connect(
            "clicked", self._start_scan
        )
        self.widget_tree.get_object("calibration_maximize").connect(
            "clicked", self._on_maximize
        )
        self.widget_tree.get_object("calibration_automatic").connect(
            "clicked", self._guess_scan_borders
        )

        drawing_area = self.widget_tree.get_object("calibration_area")
        self.core.call_all("draw_scan_start", drawing_area)

        self.core.call_all(
            "add_setting_screen",
            settings_widget_tree,
            "calibration",
            self.widget_tree.get_object("calibration_header"),
            self.widget_tree.get_object("calibration_body"),
        )

    def complete_scanner_settings(
            self, settings_widget_tree, parent_widget_tree,
            list_scanner_promise):
        assert self.widget_tree is not None

        def set_sensitive():
            dev_id = self.core.call_success("config_get", "scanner_dev_id")
            parent_widget_tree.get_object("scanner_calibration").set_sensitive(
                True if dev_id is not None and dev_id != "" else False
            )

        set_sensitive()
        self.core.call_all(
            "config_add_observer", "scanner_dev_id", set_sensitive
        )

        parent_widget_tree.get_object("scanner_calibration").connect(
            "clicked", self.display_calibration_screen, settings_widget_tree
        )

    def on_settings_closed(self, settings_widget_tree):
        drawing_area = self.widget_tree.get_object("calibration_area")
        self.core.call_all("draw_scan_stop", drawing_area)
        if self.value_changed_connect_id is not None:
            drawing_area.disconnect(self.value_changed_connect_id)
            self.value_changed_connect_id = None
        if self.size_allocate_connect_id is not None:
            self.widget_tree.get_object("calibration_scroll").disconnect(
                self.size_allocate_connect_id
            )
            self.size_allocate_connect_id = None

    def display_calibration_screen(self, *args, **kwargs):
        LOGGER.info("Switching to calibration screen")
        self.core.call_all(
            "show_setting_screen", self.settings_widget_tree, "calibration"
        )

        sources = self.widget_tree.get_object("calibration_sources")
        sources.clear()
        sources.append(("", _("Loading ...")))

        combobox = self.widget_tree.get_object("calibration_source")
        combobox.set_active(0)
        combobox.set_sensitive(False)

        buttons = [
            'calibration_automatic',
            'calibration_maximize',
            'calibration_scan',
        ]
        for button in buttons:
            self.widget_tree.get_object(button).set_sensitive(False)

        self.core.call_all(
            "on_zoomable_widget_new",
            self.widget_tree.get_object("calibration_scroll"),
            self.widget_tree.get_object("calibration_scale_adjustment")
        )

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_busy",)
        )
        promise = promise.then(
            # drop the return value of call_all()
            lambda *args, **kwargs: None
        )
        promise = promise.then(self.core.call_success(
            "scan_get_scanner_promise"
        ))
        promise = promise.then(self._show_sources)
        promise = promise.then(openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_idle",)
        ))
        self.core.call_success("scan_schedule", promise)

    def _show_sources(self, dev=None):
        sources = []
        if dev is not None:
            sources = [
                (
                    source.get_name(),
                    self.core.call_success(
                        "i18n_scanner_source", source.get_name()
                    ),
                )
                for source in dev.dev.get_children()
            ]
        LOGGER.info("Found %d sources", len(sources))

        sources_widget = self.widget_tree.get_object("calibration_sources")
        sources_widget.clear()
        for src in sources:
            LOGGER.info("Source: %s ; %s", src[0], src[1])
            sources_widget.append(src)

        combobox = self.widget_tree.get_object("calibration_source")
        combobox.set_active(0)
        combobox.set_sensitive(True)

        self.widget_tree.get_object("calibration_scan").set_sensitive(True)

    def hide_calibration_screen(self, *args, **kwargs):
        LOGGER.info("Switching back to settings")
        self.core.call_all(
            "show_setting_screen", self.settings_widget_tree, "main"
        )
        self.core.call_all(
            "on_zoomable_widget_destroy",
            self.widget_tree.get_object("calibration_area"),
            self.widget_tree.get_object("calibration_scale_adjustment")
        )

    def _start_scan(self, button):
        combobox = self.widget_tree.get_object("calibration_source")
        source_idx = combobox.get_active()
        sources = self.widget_tree.get_object("calibration_sources")
        source = sources[source_idx][0]

        LOGGER.info("Starting calibration scan on %s ...", source)

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_busy",)
        )
        promise = promise.then(
            # drop the return value of call_all()
            lambda *args, **kwargs: None
        )
        # calibration is always done at 75 DPI
        promise = promise.then(
            self.core.call_success(
                "scan_promise", source_id=source, resolution=75
            )[1]
        )
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, self._scan
        ))
        promise = promise.then(self._on_scan_end)
        promise = promise.then(openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_idle",)
        ))
        self.core.call_success("scan_schedule", promise)
        self.widget_tree.get_object("calibration_scan").set_sensitive(False)

    def _scan(self, args):
        (source, scan_id, img_generator) = args
        # just unroll the image generator.
        # --> we catch the content using the the on_scan_XXXX callbacks.
        img = None
        for (idx, img) in enumerate(img_generator):
            LOGGER.info("Page %d scanned: %s", idx, img.size)
        return img

    def on_scan_page_start(self, scan_id, page_nb, scan_params):
        (self.scan_width, self.scan_height) = (
            paperwork_backend.cairo.pillow.limit_img_size((
                scan_params.get_width(),
                scan_params.get_height(),
            ))
        )
        if self.widget_tree is None:
            return
        self._update_calibration_area_size_based_on_scroll()
        self._update_calibration_scroll_area_size()

    def _update_calibration_area_size_based_on_scroll(self, *args, **kwargs):
        scroll = self.widget_tree.get_object("calibration_scroll")
        widget_height = scroll.get_allocated_height()
        factor = widget_height / self.scan_height
        self.widget_tree.get_object("calibration_scale_adjustment").set_value(
            factor
        )
        # signal 'value-changed' will trigger the update and redraw
        self.widget_tree.get_object("calibration_scale_adjustment").set_lower(
            factor
        )

    def _update_calibration_area_size_based_on_scale(self, *args, **kwargs):
        adj = self.widget_tree.get_object(
            "calibration_scale_adjustment"
        )
        factor = adj.get_value()
        LOGGER.debug(
            "Scale: %f < %f < %f", adj.get_lower(), factor, adj.get_upper()
        )

        widget_width = int(self.scan_width * factor)
        widget_height = int(self.scan_width * factor)

        LOGGER.debug(
            "Calibratrion widget size: (%d, %d)", widget_width, widget_height
        )
        self.widget_tree.get_object("calibration_area").set_size_request(
            widget_width, widget_height
        )
        self.widget_width = widget_width

    def _update_calibration_scroll_area_size(self, *args, **kwargs):
        # scroll area must have the same proportion than the scanned image
        scroll = self.widget_tree.get_object("calibration_scroll")
        widget_height = scroll.get_allocated_height()
        if self.scan_height <= 0:
            ratio = 1.0 / 1.414
        else:
            ratio = self.scan_width / self.scan_height
        widget_width = max(widget_height * ratio, 300)
        LOGGER.debug(
            "Calibration scroll window size: (%d, %d)",
            widget_width, widget_height
        )
        scroll.set_size_request(widget_width, -1)

        if self.scan_height > 0:
            factor = widget_height / self.scan_height
            adj = self.widget_tree.get_object("calibration_scale_adjustment")
            if adj.get_value() < factor:
                adj.set_value(factor)
            adj.set_lower(factor)

    def _on_scan_end(self, scan_img=None):
        if self.widget_tree is not None:
            buttons = [
                'calibration_automatic',
                'calibration_maximize',
                'calibration_scan'
            ]
            for button in buttons:
                self.widget_tree.get_object(button).set_sensitive(True)

            drawing_area = self.widget_tree.get_object("calibration_area")
            self.core.call_all("draw_scan_stop", drawing_area)

        if scan_img is None:
            LOGGER.info("No page scanned. Can't do calibration")
            return

        (self.scan_width, self.scan_height) = (
            paperwork_backend.cairo.pillow.limit_img_size(scan_img.size)
        )
        self.scan_img = scan_img

        LOGGER.info("Calibration scan ready")

        if self.core.call_success("config_get", "scanner_calibration") is None:
            # Put the frame a little bit inside the image to make the
            # corner handles more visible
            calibration = [
                min(50, self.scan_width),
                min(50, self.scan_height),
                max(self.scan_width - 50, 0),
                max(self.scan_height - 50, 0),
            ]
            LOGGER.info("Setting default calibration area: %s", calibration)
            self.core.call_all(
                "config_put", "scanner_calibration", calibration
            )

        self.core.call_all("draw_pillow_start", drawing_area, scan_img)
        self.core.call_all(
            "draw_calibration_start", drawing_area,
            (self.scan_width, self.scan_height)
        )

    def _on_maximize(self, button):
        if self.scan_height <= 0 or self.scan_width <= 0:
            return
        self.core.call_all(
            "config_put", "scanner_calibration",
            [0, 0, self.scan_width, self.scan_height]
        )

    def _guess_scan_borders(self, button=None):
        if self.scan_img is None:
            return

        LOGGER.info("Guessing scan borders")

        def find_scan_borders(scan_img):
            frame = pillowfight.find_scan_borders(scan_img)
            return frame

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_busy",)
        )
        promise = promise.then(lambda *args, **kwargs: self.scan_img)
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, find_scan_borders
        ))
        promise = promise.then(
            lambda frame: self.core.call_all(
                "config_put", "scanner_calibration", list(frame)
            )
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.core.call_all, "on_idle")
        promise.schedule()

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.widget_tree is None:
            return
        if self.settings_widget_tree is None:
            return

        body = self.widget_tree.get_object("calibration_body")
        if body is None:
            return
        self.core.call_success(
            "screenshot_snap_widget", body,
            self.core.call_success(
                "fs_join", out_dir, "settings_calibration_dialog.png"
            ),
            margins=(100, 100, 100, 100)
        )



import getpass
import logging
import os
import socket
import urllib.parse

from . import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self):
        self.replacements = (
            (os.path.expanduser("~"), "###HOME_DIR###"),
            (urllib.parse.quote(os.path.expanduser("~")), "###HOME_DIR###"),
            (getpass.getuser(), "###USER_NAME###"),
            (urllib.parse.quote(getpass.getuser()), "###USER_NAME###"),
            (socket.gethostname(), "###HOST_NAME###"),
            (urllib.parse.quote(socket.gethostname()), "###HOST_NAME###"),
        )

    def get_interfaces(self):
        return ['censor']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
        ]

    def censor_string(self, string):
        """
        Rewrite a string by censoring anything close to personal information
        (username, home directory path, host name, ...).
        """
        for r in self.replacements:
            string = string.replace(*r)
        return string

    def censor_txt_file(
            self, input_url, output_url=None, tmp_on_disk=False):
        """
        Rewrite a file but censor anything close to personal information
        (username, home directory path, host name, ...).

        Arguments:
        - file_input_url: file to censor
        - file_output_url: output file. If None, a temporary file will be
          created
        - tmp_on_disk: only used if file_output_url is None. If true,
          the temporary file will be written on disk. If false, we may
          return a memory:// URI (won't work outside of Paperwork).

        Returns:
          censored file URL
        """
        if output_url is not None:
            fd_out = self.core.call_success("fs_open", output_url, 'w')
        else:
            basename = self.core.call_success("fs_basename", input_url)
            (output_url, fd_out) = self.core.call_success(
                "fs_mktemp", prefix="censored_", suffix="_" + basename,
                mode="w", on_disk=tmp_on_disk
            )

        with fd_out:
            with self.core.call_success("fs_open", input_url, 'r') as fd_in:
                while True:
                    line = fd_in.readline()
                    if line == '':
                        break
                    line = self.censor_string(line)
                    fd_out.write(line)

        return output_url



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def __init__(self):
        super().__init__()
        self.visible = False

    def get_interfaces(self):
        return [
            'gtk_pageview_boxes_all',
            'gtk_pageview_boxes_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_pageview',
                'defaults': ['paperwork_gtk.mainwindow.docview.pageview'],
            },
            {
                'interface': 'gtk_pageview_boxes',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageview.boxes'
                ],
            },
        ]

    def set_all_boxes_visible(self, visible):
        self.visible = visible
        self.core.call_all("pageview_refresh_all")

    def on_page_boxes_loaded(self, page, boxes, spatial_index):
        page.refresh()

    def on_page_draw(self, cairo_ctx, page):
        if not self.visible:
            return

        boxes = self.core.call_success(
            "pageview_get_boxes_by_id", page.doc_id, page.page_idx
        )
        if boxes is None:
            return

        for line_box in boxes:
            for word_box in line_box.word_boxes:
                self.core.call_all(
                    "page_draw_box",
                    cairo_ctx, page, word_box.position,
                    (0.0, 0.0, 1.0), border_width=1
                )



"""
Mock implementation of the plugin interface 'fs'.
Useful for tests only.
"""

import io
import logging
import os

from . import CommonFsPluginBase
from . import memory

LOGGER = logging.getLogger(__name__)


class FakeFileAdapter(io.RawIOBase):
    def __init__(self, fs_plugin, path, content, mode='r'):
        super().__init__()
        self.fs_plugin = fs_plugin
        self.path = path
        self.content = content if 'w' not in mode else None
        self.pos = len(self.content) if 'a' in mode else 0
        self.mode = mode

        if self.content is None:
            self.content = b'' if 'b' in mode else ''

    def readable(self):
        return True

    def writable(self):
        return 'w' in self.mode or 'a' in self.mode

    def read(self, size=-1):
        if size < 0:
            r = len(self.content[self.pos:])
        else:
            r = min(size, len(self.content[self.pos:]))
        out = self.content[self.pos:self.pos + r]
        self.pos += r
        return out

    def readall(self):
        return self.content

    def readinto(self, b):
        raise NotImplementedError()

    def readline(self, size=-1):
        r = self.content[self.pos:]
        r = r.split("\n", 1)
        self.pos += len(r[0]) + 1
        return r[0] + "\n"

    def readlines(self, hint=-1):
        self.pos = len(self.content)
        return [
            r + "\n"
            for r in self.content.split("\n")
        ]

    def seek(self, offset, whence=os.SEEK_SET):
        if whence == os.SEEK_CUR:
            self.pos += offset
        elif whence == os.SEEK_SET:
            self.pos = offset
        else:
            raise NotImplementedError()

    def seekable(self):
        return True

    def tell(self):
        return self.pos

    def flush(self):
        pass

    def truncate(self, size=None):
        raise NotImplementedError()

    def isatty(self):
        return False

    def write(self, b):
        self.content = self.content[:self.pos]
        self.content += b
        self.pos += len(b)
        return len(b)

    def writelines(self, lines):
        raise NotImplementedError()

    def close(self):
        if 'a' in self.mode or 'w' in self.mode:
            d = self.fs_plugin.fs
            for p in self.path[:-1]:
                d = d[p]
            d[self.path[-1]] = self.content
        self.flush()
        super().close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class Plugin(CommonFsPluginBase):
    def __init__(self, fs=None):
        """
        Test implementation.

        `fs` should have the following format:

        {
            'base_dir': {
                'sub_dir_a': {},
                'sub_dir_b': {
                    'file_a': "content_file_a",
                    'file_b': "content_file_b",
                }
            }
        }
        """
        super().__init__()
        if fs is None:
            self.fs = {}
        else:
            self.fs = fs

    @staticmethod
    def _get_path(url):
        assert url.lower().startswith("file:///")
        url = url[len('file:///'):]
        return url.split('/')

    def fs_open(self, url, mode='r', **kwargs):
        path = self._get_path(url)

        f = self.fs
        for p in path[:-1]:
            f = f[p]
        if path[-1] in f:
            f = f[path[-1]]
            assert isinstance(f, str) or isinstance(f, bytes)
        elif 'b' in mode:
            f = b""
        else:
            f = ""
        return FakeFileAdapter(self, path, f, mode)

    def fs_exists(self, url):
        path = self._get_path(url)
        f = self.fs
        for p in path:
            if p not in f:
                return None
            f = f[p]
        return True

    def fs_listdir(self, url):
        path = self._get_path(url)
        f = self.fs
        for p in path:
            try:
                f = f[p]
            except KeyError:
                return None
        assert isinstance(f, dict)
        return [url + "/" + k for k in f.keys()]

    def fs_rename(self, old_url, new_url):
        old_path = self._get_path(old_url)
        old_dir = self.fs
        for p in old_path[:-1]:
            old_dir = old_dir[p]

        new_path = self._get_path(new_url)
        new_dir = self.fs
        for p in new_path[:-1]:
            new_dir = new_dir[p]

        old_file = old_dir.pop(old_path[-1])
        new_dir[new_path[-1]] = old_file

    def fs_unlink(self, url, **kwargs):
        self.fs_rm_rf(url)

    def fs_rm_rf(self, url, **kwargs):
        path = self._get_path(url)
        f = self.fs
        for p in path[:-1]:
            if p not in f:
                return None
            f = f[p]
        assert isinstance(f, dict)
        filename = url.split("/")[-1]
        if filename in f:
            f.pop(filename)
            return True
        return None

    def fs_get_mtime(self, url):
        return 0

    def fs_getsize(self, url):
        path = self._get_path(url)
        f = self.fs
        for p in path:
            f = f[p]
        return len(f)

    def fs_isdir(self, url):
        path = self._get_path(url)
        f = self.fs
        for p in path:
            if p not in f:
                return None
            f = f[p]
        if isinstance(f, dict):
            return True
        return None

    def fs_copy(self, old_url, new_url):
        raise NotImplementedError()

    def fs_mkdir_p(self, url):
        path = self._get_path(url)
        f = self.fs
        for p in path[:-1]:
            if p not in f:
                f[p] = {}
            f = f[p]
        name = url.split("/")[-1]
        if name not in f:
            f[name] = {}
        return True

    def fs_recurse(self, parent_uri, dir_included=False):
        raise NotImplementedError()

    def fs_hide(self, uri):
        pass

    def fs_get_mime(self, uri):
        pass

    def fs_iswritable(self, url):
        return True

    def fs_mktemp(self, prefix=None, suffix=None, mode='w+b', **kwargs):
        name = "file://tmp/temporary_file" + suffix
        return (name, memory.MemoryFileAdapter(self, name, mode))



"""
Wraps a pageview in a GtkOverlay + GtkSpinner. Displays the spinner
on top of the pageview when another plugin is working on the page.
"""

import logging

import openpaperwork_core

try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

    # workaround so chkdeps can still be called
    class GObject(object):
        TYPE_BOOLEAN = 0

        class SignalFlags(object):
            RUN_LAST = 0

        class GObject(object):
            pass


LOGGER = logging.getLogger(__name__)


class PageWrapper(GObject.GObject):
    __gsignals__ = {
        'getting_size': (GObject.SignalFlags.RUN_LAST, None, ()),
        'size_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'img_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'visibility_changed': (GObject.SignalFlags.RUN_LAST, None, (
            GObject.TYPE_BOOLEAN,
        )),
    }

    def __init__(self, plugin, page):
        super().__init__()
        self.plugin = plugin
        self.page = page
        self.page_idx = page.page_idx
        self.busy = False

        self.widget_tree = self.plugin.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageprocessing",
            "pageprocessing.glade",
        )
        self.widget = self.widget_tree.get_object("pageprocessing_overlay")

        self.widget.add(page.widget)

        page.connect("getting_size", lambda p: self.emit('getting_size'))
        page.connect("size_obtained", lambda p: self.emit("size_obtained"))
        page.connect("img_obtained", lambda p: self.emit("img_obtained"))
        page.connect(
            "visibility_changed",
            lambda p, v: self.emit("visibility_changed", v)
        )

        page.widget.connect("draw", self._on_draw)

    def load(self):
        self.page.load()

    def close(self):
        self.page.close()

    def get_zoom(self):
        return self.page.get_zoom()

    def set_zoom(self, zoom):
        return self.page.set_zoom(zoom)

    def get_full_size(self):
        return self.page.get_full_size()

    def get_size(self):
        return self.page.get_size()

    def resize(self):
        return self.page.resize()

    def refresh(self, reload=False):
        return self.page.refresh(reload=reload)

    def set_visible(self, visible):
        return self.page.set_visible(visible)

    def get_visible(self):
        return self.page.get_visible()

    def _on_draw(self, overlay, cairo_ctx):
        if not self.busy:
            return

        w = overlay.get_allocated_width()
        h = overlay.get_allocated_height()

        cairo_ctx.set_source_rgba(0.5, 0.5, 0.5, 0.5)
        cairo_ctx.rectangle(0, 0, w, h)
        cairo_ctx.fill()

    def on_page_modification_start(self):
        self.busy = True
        spinner = self.widget_tree.get_object("pageprocessing_spinner")
        spinner.set_visible(True)
        spinner.start()

    def on_page_modification_end(self):
        self.busy = False
        spinner = self.widget_tree.get_object("pageprocessing_spinner")
        spinner.set_visible(False)
        spinner.stop()
        self.page.refresh(reload=True)


if GLIB_AVAILABLE:
    GObject.type_register(PageWrapper)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -1000

    def __init__(self):
        super().__init__()
        self.wrappers = []
        self.active_doc_id = None
        self.active_pages = set()

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_pageview',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_close(self):
        self.wrappers = []

    def doc_open_components(self, out: list, doc_id, doc_url):
        if doc_id != self.active_doc_id:
            self.active_doc_id = doc_id
            self.active_pages = set()

        self.doc_close()

        # instantiate PageWrapper objects, and replace the pages in the
        # list 'out' by those wrappers.
        self.wrappers = [
            PageWrapper(self, page)
            for (visible, page) in out
        ]
        for page_idx in range(0, len(out)):
            out[page_idx] = (out[page_idx][0], self.wrappers[page_idx])
            if page_idx in self.active_pages:
                out[page_idx][1].on_page_modification_start()

    def on_page_modification_start(self, doc_id, page_idx):
        if doc_id != self.active_doc_id:
            return
        self.active_pages.add(page_idx)
        if page_idx >= len(self.wrappers):
            return
        self.wrappers[page_idx].on_page_modification_start()

    def on_page_modification_end(self, doc_id, page_idx):
        if doc_id != self.active_doc_id:
            return
        if page_idx in self.active_pages:
            self.active_pages.remove(page_idx)
        if page_idx >= len(self.wrappers):
            return
        self.wrappers[page_idx].on_page_modification_end()



import openpaperwork_core


class SpatialIndexer(object):
    def __init__(self, boxes):
        self.boxes = list(boxes)

    def get_boxes(self, pt_x, pt_y):
        for (pos, obj) in self.boxes:
            if pos[0][0] > pt_x:
                continue
            if pos[0][1] > pt_y:
                continue
            if pos[1][0] < pt_x:
                continue
            if pos[1][1] < pt_y:
                continue
            yield (pos, obj)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'spatial_index'
        ]

    def spatial_indexer_get(self, boxes):
        return SpatialIndexer(boxes)



def parse_page_list(args):
    if not hasattr(args, 'pages'):
        return None
    if args.pages is None or args.pages == "":
        return None

    if "-" in args.pages:
        pages = args.pages.split("-", 1)
        return range(
            int(pages[0]) - 1,
            int(pages[1])
        )
    else:
        return [
            (int(p) - 1) for p in args.pages.split(",")
            if int(p) >= 1
        ]



import logging

import openpaperwork_core
import openpaperwork_core.deps

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -1000

    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return [
            'gtk_settings',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'stats_post',
                'defaults': ['openpaperwork_core.beacon.stats'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
        ]

    def init(self, core):
        super().init(core)

    def complete_settings(self, global_widget_tree):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.settings", "stats.glade"
        )

        active = self.core.call_success("config_get", "send_statistics")
        LOGGER.info("Statistics state: %s", active)

        button = widget_tree.get_object("stats_state")
        button.set_active(active)
        button.connect("notify::active", self._on_stats_state_changed)

        button = widget_tree.get_object("stats_infos")
        details = widget_tree.get_object("stats_details")
        button.connect("clicked", self._on_info_button, details)

        self.core.call_success(
            "add_setting_to_dialog", global_widget_tree,
            _("Help Improve Paperwork"),
            [widget_tree.get_object("stats")]
        )

    def _on_info_button(self, info_button, details):
        details.set_visible(not details.get_visible())

    def _on_stats_state_changed(self, switch, _):
        state = switch.get_active()
        LOGGER.info("Setting stats state to %s", state)
        self.core.call_all("config_put", "send_statistics", state)



import hashlib
import logging
import os
import pathlib
import urllib
import urllib.parse

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class CommonFsPluginBase(openpaperwork_core.PluginBase):
    def __init__(self):
        """
        Should be used only by sub-classes
        """
        super().__init__()

    def get_interfaces(self):
        return ['fs']

    @staticmethod
    def fs_safe(uri):
        """
        Make sure the specified URI is actually an URI and not a Unix path.
        Returns:
            - An URI
        """
        if isinstance(uri, pathlib.Path):
            return uri.as_uri()
        LOGGER.debug("safe: %s", uri)
        if uri[:2] == "\\\\" or "://" in uri:
            LOGGER.debug("safe: --> %s", uri)
            return uri
        return pathlib.Path(uri).as_uri()

    @staticmethod
    def fs_unsafe(uri):
        """
        Turn an URI into an Unix path, whenever possible.
        Shouldn't be used at all.
        """
        LOGGER.debug("unsafe: %s", uri)
        if "://" not in uri and uri[:2] != "\\\\":
            LOGGER.debug("unsafe: --> %s", uri)
            return uri
        if not uri.startswith("file://"):
            LOGGER.debug("unsafe: --> EXC")
            raise Exception("TARGET URI SHOULD BE A LOCAL FILE")
        uri = uri[len("file://"):]
        if os.name == 'nt' and uri[0] == '/':
            # for some reason, URIs on Windows start with
            # "file:///C:\..." instead of "file://C:\..."
            uri = uri[1:]
        uri = urllib.parse.unquote(uri)
        return uri

    # It seems urllib.parse.urljoin doesn't work any protocol. For instance,
    # it doesn't play nice with smb:// and davs://
    URLLIB_SUPPORTED_PROTOCOLS = {"http", "https", "file", "ftp"}

    def fs_join(self, base, url):
        if not base.endswith("/"):
            base += "/"
        protocol = base.split("://", 1)[0].lower()
        if protocol in self.URLLIB_SUPPORTED_PROTOCOLS:
            return urllib.parse.urljoin(base, url)
        return base + url

    def fs_basename(self, url):
        parsed_url = urllib.parse.urlparse(url)
        path = parsed_url.path
        if path == "":
            path = parsed_url.hostname
        if path is None:
            return None
        basename = os.path.basename(path)
        # Base name can be safely unquoted
        return urllib.parse.unquote(basename)

    def fs_dirname(self, url):
        # dir name should not be unquoted. It could mess up the URI
        return os.path.dirname(url)

    def fs_hash(self, url):
        with self.core.call_success("fs_open", url, 'rb') as fd:
            content = fd.read()
        return int(hashlib.sha256(content).hexdigest(), 16)

    def fs_copy(self, origin_url, dest_url):
        """
        default generic implementation
        """
        with open(origin_url, 'rb') as fd:
            content = fd.read()
        with open(dest_url, 'wb') as fd:
            fd.write(content)
        return True



import openpaperwork_core

import PIL
import PIL.Image


from . import AbstractImgEditor


class RotationImgEditor(AbstractImgEditor):
    def __init__(self, angle):
        self.angle = angle % 360

    def transform(self, img, preview=False):
        # Pillow operates counter-clockwise, we operate clockwise.
        if self.angle == 0:
            return img
        transform_enum = getattr(PIL.Image, 'Transpose', PIL.Image)
        angle = {
            90: transform_enum.ROTATE_90,
            180: transform_enum.ROTATE_180,
            270: transform_enum.ROTATE_270,
        }[self.angle]
        return img.transpose(angle)

    def _transform_frame(self, img_size, frame, transform_pt):
        frame = (
            transform_pt(img_size, (frame[0], frame[1])),
            transform_pt(img_size, (frame[2], frame[3])),
        )
        return (
            min(frame[0][0], frame[1][0]),
            min(frame[0][1], frame[1][1]),
            max(frame[0][0], frame[1][0]),
            max(frame[0][1], frame[1][1]),
        )

    def transform_frame(self, img_size, frame):
        return self._transform_frame(img_size, frame, self.transform_point)

    def untransform_frame(self, img_size, frame):
        return self._transform_frame(img_size, frame, self.untransform_point)

    def _transform_pt(self, img_size, pt, angle):
        r = {
            0: pt,
            90: (pt[1], img_size[0] - pt[0]),
            270: (img_size[1] - pt[1], pt[0]),
            180: (img_size[0] - pt[0], img_size[1] - pt[1]),
        }[angle]
        return r

    def transform_point(self, img_size, pt):
        return self._transform_pt(img_size, pt, self.angle)

    def untransform_point(self, img_size, pt):
        # we are given the image size before we apply our transformation
        # but here we want to go the other way around
        img_size = {
            0: img_size,
            90: (img_size[1], img_size[0]),
            270: (img_size[1], img_size[0]),
            180: img_size,
        }[self.angle]
        angle = ((-1 * self.angle) % 360)
        return self._transform_pt(img_size, pt, angle)


class Plugin(openpaperwork_core.PluginBase):
    NAME = 'rotation'

    def get_interfaces(self):
        return ['img_editor']

    def img_editor_get_names(self, out: list):
        out.append(self.NAME)

    def img_editor_get(self, name, *args, **kwargs):
        if name != self.NAME:
            return None
        angle = kwargs.pop('angle')
        return RotationImgEditor(angle)

    def img_editor_set(self, inout: list, name, *args, **kwargs):
        if name != self.NAME:
            return None
        angle = kwargs.pop('angle')
        c = RotationImgEditor(angle)
        # Check if we already have a RotationImgEditor in the list.
        # If so, update it instead of adding another one
        # Only if its the latest element in the list -> otherwise we may
        # mess up the cropping frame
        if len(inout) > 0 and inout[-1] == c:
            inout[-1].angle += angle
            inout[-1].angle %= 360
        else:
            inout.append(c)

    def img_editor_unset(self, inout: list, name):
        if name != self.NAME:
            return None
        c = RotationImgEditor(0)
        while c in inout:
            inout.remove(c)



import logging
import threading
import time

from .. import PluginBase


LOGGER = logging.getLogger(__name__)

MIN_TIME_MS = 200


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.pending = {}

    def get_interfaces(self):
        return ['perfcheck']

    def on_perfcheck_start(self, task_name):
        k = (task_name, threading.get_ident())
        self.pending[k] = time.time()

    def on_perfcheck_stop(self, task_name, **extras):
        stop = time.time()
        k = (task_name, threading.get_ident())
        start = self.pending.pop(k)
        if (stop - start) * 1000 >= MIN_TIME_MS:
            LOGGER.warning(
                "Task '%s' took %dms (> %dms) ! (%s)",
                task_name, (stop - start) * 1000, MIN_TIME_MS, extras
            )



from paperwork_backend import _
from paperwork_backend.docexport import (
    AbstractExportPipe,
    AbstractExportPipePlugin,
    ExportData,
    ExportDataType
)


class AutomaticPdfExportPipe(AbstractExportPipe):
    """
    Composite pipe that generates PDF using the original PDF if possible.
    It automatically select between:
    - copying the original PDF file (if available and if the doc is unmodified)
    - exporting a generated PDF
    """
    def __init__(self, core):
        super().__init__(
            name="automatic_pdf",
            input_type=ExportDataType.DOCUMENT,
            output_type=ExportDataType.OUTPUT_URL_FILE,
        )
        self.core = core
        self.can_change_quality = True
        self.can_change_page_format = True

        self.pipes_original_pdf = [
            self.core.call_success("export_get_pipe_by_name", "unmodified_pdf")
        ]
        assert None not in self.pipes_original_pdf
        self.pipes_generated_pdf = [
            self.core.call_success("export_get_pipe_by_name", "doc_to_pages"),
            self.core.call_success("export_get_pipe_by_name", "img_boxes"),
            self.core.call_success("export_get_pipe_by_name", "generated_pdf"),
        ]
        assert None not in self.pipes_generated_pdf

    def can_export_doc(self, doc_url):
        return True

    def export(self, input_data, result="final", target_file_url=None):
        assert input_data.dtype == ExportDataType.DOCUMENT_SET
        list_docs = input_data.iter(ExportDataType.DOCUMENT)
        out_files = []
        for (idx, (doc_set, doc)) in enumerate(list_docs):
            assert doc.dtype == ExportDataType.DOCUMENT
            (doc_id, doc_url) = doc.data

            doc_input = ExportData.build_doc(doc_id, doc_url)

            out_url = target_file_url
            if idx != 0:
                # automatic names files slightly differently: we assume
                # the user is doing a bulk export.
                out_url = target_file_url.rsplit(".", 1)
                out_url = f"{out_url[0]}_{doc_id}.{out_url[1]}"

            pdf_url = self.core.call_success(
                "doc_get_pdf_url_by_url", doc_url, allow_mapped=False
            )
            if pdf_url is not None:
                pipes = self.pipes_original_pdf
            else:
                pipes = self.pipes_generated_pdf

            r = doc_input
            for pipe in pipes:
                if pipe.can_change_quality:
                    pipe.set_quality(self.quality)
                if pipe.can_change_page_format:
                    pipe.set_page_format(self.page_format)
                r = pipe.export(r, result, out_url)
            out_files += r

        return out_files

    def get_output_mime(self):
        return ("application/pdf", ("pdf",))

    def __str__(self):
        return _("Automatic PDF export (original PDF if possible, generated one else)")


class Plugin(AbstractExportPipePlugin):
    def get_deps(self):
        return super().get_deps() + [
            {
                "interface": "export_pipes_generic",
                "defaults": ["paperwork_backend.docexport.generic"],
            },
            {
                "interface": "export_pipes_img",
                "defaults": ["paperwork_backend.docexport.img"],
            },
            {
                "interface": "export_pipes_pdf",
                "defaults": ["paperwork_backend.docexport.pdf"],
            },
        ]

    def init(self, core):
        super().init(core)
        self.pipes = [
            AutomaticPdfExportPipe(core),
        ]



import logging

import openpaperwork_core
import openpaperwork_core.promise


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    RECOMMENDED = 300
    MODES = [
        ('radioColor', 'Color'),
        ('radioGrayscale', 'Gray'),
        ('radioLineart', 'Lineart'),
    ]

    def get_interfaces(self):
        return [
            'gtk_settings_scanner_setting',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_settings_scanner',
                'defaults': ['paperwork_gtk.settings.scanner.settings'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
        ]

    def complete_scanner_settings(
            self, global_widget_tree, parent_widget_tree,
            list_scanner_promise):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.settings.scanner",
            "popover_mode.glade"
        )

        active = self.core.call_success("config_get", "scanner_mode")
        for (widget, mode) in self.MODES:
            if mode == active:
                widget_tree.get_object(widget).set_active(True)

        for (widget, mode) in self.MODES:
            widget_tree.get_object(widget).connect(
                "toggled", self._on_toggle,
                widget_tree, mode
            )

        selector = widget_tree.get_object("selector")

        # WORKAROUND(Jflesch): set_sensitive() doesn't appear to work on
        # GtkMenuButton --> we have to play with set_popover()

        def reset_popover():
            dev_id = self.core.call_success("config_get", "scanner_dev_id")
            parent_widget_tree.get_object("scanner_mode").set_popover(
                selector if dev_id is not None and dev_id != "" else None
            )

        reset_popover()
        self.core.call_all(
            "config_add_observer", "scanner_dev_id", reset_popover
        )

    def _on_toggle(self, checkbox, widget_tree, mode):
        LOGGER.info("Selected mode: %s", mode)
        widget_tree.get_object("selector").popdown()
        self.core.call_success("config_put", "scanner_mode", mode)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "page_move_to_doc"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = None
        self.active_page_idx = -1
        self.waiting_for_target_doc = False

    def get_interfaces(self):
        return [
            'action',
            'action_page',
            'action_page_move_to_doc',
            'chkdeps',
            'doc_open',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_search_field',
                'defaults': ['paperwork_gtk.mainwindow.search.field'],
            },
            {
                'interface': 'pages',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.img_overlay',
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return

        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.actions.page.move_to_doc", "move_to_doc.glade"
        )
        if widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return
        widget_tree.get_object("move_to_doc_cancel").connect(
            "clicked", self._cancel_move_to_doc
        )
        self.core.call_all(
            "mainwindow_add",
            "right", "move_to_doc",
            -9999999999,
            widget_tree.get_object("move_to_doc_header"),
            widget_tree.get_object("move_to_doc"),
        )

        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._start_move_to_doc)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_open(self, doc_id, doc_url):
        new_active_doc = (doc_id, doc_url)
        if self.waiting_for_target_doc:
            self.waiting_for_target_doc = False
            self._move_to_doc(new_active_doc)
        self.active_doc = new_active_doc

    def doc_close(self):
        self.active_doc = None

    def on_page_shown(self, page_idx):
        self.active_page_idx = page_idx

    def _start_move_to_doc(self, *args, **kwargs):
        assert self.active_doc is not None
        if self.active_page_idx < 0:
            return
        LOGGER.info("Starting 'move to doc' process")
        self.core.call_all("mainwindow_show", "right", "move_to_doc")
        self.core.call_all("mainwindow_show", "left", "doclist")
        self.waiting_for_target_doc = True

    def _run_transaction(self, original_doc, target_doc):
        nb_remaining_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", original_doc[1]
        )
        if nb_remaining_pages is None:
            nb_remaining_pages = 0

        self.core.call_success(
            "transaction_simple_promise",
            [
                ('upd', target_doc[0]),
                (
                    'upd' if nb_remaining_pages > 0 else 'del',
                    original_doc[0]
                ),
            ]
        )

    def _move_to_doc(self, target_doc):
        promise = openpaperwork_core.promise.Promise(self.core)
        promise = promise.then(
            self.core.call_all,
            "page_move_by_url",
            self.active_doc[1], self.active_page_idx,
            target_doc[1], 0
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_one,
            "mainloop_schedule", self.core.call_all,
            "doc_reload", *self.active_doc
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_one,
            "mainloop_schedule", self.core.call_all,
            "doc_reload", *target_doc
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_all,
            "search_update_document_list"
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self._run_transaction, self.active_doc, target_doc
        )

        self.core.call_success("transaction_schedule", promise)

    def _cancel_move_to_doc(self, *args, **kwargs):
        LOGGER.info("Canceling 'move to doc' process")
        self.core.call_all("mainwindow_back", "right")
        self.waiting_for_target_doc = False



"""
When drag'n'dropping, enables the dragging-from-pages part.
"""

import logging

try:
    import gi
    gi.require_version('Gdk', '3.0')
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gdk
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)

TARGET_ENTRY_URI = 0


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -10000

    def __init__(self):
        super().__init__()
        self.pages = []
        self.active_doc = None
        self.enabled = False

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_pageview',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'page_img',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.pdf',
                ],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def _enable_drag(self, widget, doc_id, doc_url, page_idx):
        widget.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, [], Gdk.DragAction.MOVE
        )

        targets = Gtk.TargetList.new([])
        targets.add_uri_targets(TARGET_ENTRY_URI)
        widget.drag_source_set_target_list(targets)

        widget.drag_source_set_icon_name("document-send-symbolic")

        widget.connect(
            "drag-data-get", self._on_drag_data_get, doc_id, doc_url, page_idx
        )

    def _disable_drag(self, widget):
        widget.drag_source_unset()

    def _on_drag_data_get(
            self, widget, drag_context, data, info, time,
            doc_id, doc_url, page_idx):
        LOGGER.info("drag_data_get(%s, p%d, type=%d)", doc_id, page_idx, info)

        if info == TARGET_ENTRY_URI:
            # drag'n'drop API allows us to provide many URIs. But if we
            # do, applications like Firefox will try to display them all,
            # even if they don't understand the URI scheme.
            # --> we cheat and use the URI target for the extra info we may
            # need in Paperwork
            img_url = self.core.call_success(
                "page_get_img_url", doc_url, page_idx
            )
            if img_url is None:
                return
            img_url = img_url.split("#", 1)[0]
            img_url += "#doc_id={}&page={}".format(doc_id, page_idx)
            LOGGER.info("Img URL: {}".format(img_url))
            data.set_uris([img_url])
            return

        assert False

    def doc_open_components(self, out: list, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)
        self.pages = [p[1] for p in out]
        if self.enabled:
            for (visible, page) in out:
                if visible:
                    self._enable_drag(
                        page.widget, doc_id, doc_url, page.page_idx
                    )

    def on_layout_change(self, layout_name):
        enabled = (layout_name == 'grid')
        if enabled == self.enabled:
            return
        if enabled:
            LOGGER.info("Layout %s --> Enabling drag(-and-drop)", layout_name)
            for page in self.pages:
                self._enable_drag(page.widget, *self.active_doc, page.page_idx)
        else:
            LOGGER.info("Layout %s --> Disabling drag(-and-drop)", layout_name)
            for page in self.pages:
                self._disable_drag(page.widget)
        self.enabled = enabled



"""
module 'http' contains all the code to communicate using HTTP+JSON with
https://www.openpaper.work/:
- update notification
- anonymous statistics
- etc
"""
import base64
import http
import http.client
import json
import logging
import os
import ssl
import urllib

from . import PluginBase
from . import promise


if os.name == "nt":
    import certifi


LOGGER = logging.getLogger(__name__)
DEFAULT_SERVER = "www.openpaper.work"
DEFAULT_PROTOCOL = "https"


class JsonHttp(object):
    def __init__(self, core, module_name):
        self.core = core
        self.user_agent = "{} {}".format(
            core.call_success("app_get_name"),
            core.call_success("app_get_version")
        )
        self.config_section_name = module_name

        settings = {
            self.config_section_name + "_protocol":
                core.call_success(
                    "config_build_simple", self.config_section_name,
                    "protocol", lambda: DEFAULT_PROTOCOL
                ),
            self.config_section_name + "_server":
                core.call_success(
                    "config_build_simple", self.config_section_name,
                    "server", lambda: DEFAULT_SERVER
                ),
        }
        for (k, setting) in settings.items():
            core.call_all("config_register", k, setting)

    def _convert(self, data):
        if isinstance(data, str):
            return data
        if isinstance(data, bytes):
            return base64.encodebytes(data).decode("utf-8").strip()
        return json.dumps(data)

    def _request(self, data, protocol, server, path):
        if protocol == "http":
            h = http.client.HTTPConnection(host=server)
        elif os.name == "nt":
            # On Windows, when frozen, we must specify a cafile
            cafile = certifi.where()
            ssl_context = ssl.create_default_context(cafile=cafile)
            h = http.client.HTTPSConnection(host=server, context=ssl_context)
        else:
            h = http.client.HTTPSConnection(host=server)

        if data is None or (isinstance(data, str) and data == ""):
            LOGGER.info("Sending GET %s/%s", server, path)
            h.request('GET', url=path, headers={'User-Agent': self.user_agent})
        else:
            body = urllib.parse.urlencode({
                k: self._convert(v)
                for (k, v) in data.items()
            })
            LOGGER.info("Sending POST %s/%s", server, path)
            h.request(
                'POST', url=path, headers={
                    "Content-type": "application/x-www-form-urlencoded",
                    "Accept": "text/plain",
                    'User-Agent': self.user_agent,
                },
                body=body
            )
        r = h.getresponse()
        reply = r.read().decode('utf-8')
        LOGGER.info("Got HTTP %s: %s - %s", r.status, r, reply)
        if r.status != http.client.OK:
            raise ConnectionError("HTTP {}: {} - {}".format(
                r.status, r, reply
            ))
        if reply[0] != '[' and reply[0] != '{':
            return reply
        return json.loads(reply)

    def get_request_promise(self, path):
        protocol = self.core.call_success(
            "config_get", self.config_section_name + "_protocol"
        )
        server = self.core.call_success(
            "config_get", self.config_section_name + "_server"
        )

        return promise.ThreadedPromise(
            self.core, self._request, args=(protocol, server, path)
        )


class Plugin(PluginBase):
    def get_interfaces(self):
        return ['http_json']

    def get_deps(self):
        return [
            {
                'interface': 'app',
                'defaults': ['openpaperwork_core.app'],
            },
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
        ]

    def http_json_get_client(self, module_name):
        return JsonHttp(self.core, module_name)
